#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "avrcore.h"

int flash_size = 8192;
uint16_t flash[8192];
int ram_size = 2048;
uint8_t ram[2048];

uSREG sreg;
uint16_t SP;
uint16_t PC;
uint8_t gpr[16];

uint8_t ram_addr_hi;
uint8_t ram_addr;

uint8_t gpior0;
uint8_t gpior1;
uint8_t gpior2;
uint8_t nvmcsr = 0x80;
uint8_t nvmcmd;
uint8_t flash_tmp;

uint8_t exbi;	// processor support extended cbi/sbi

int sleep_occ;
int wdr_occ;

int log_done;

void dump_all(int count) {
	int i,addr,n;
	int match = 0;
	printf("*** inst-count %d ***\n", count);
	printf("*** gpr dump ***\n");
	for (i=0; i<8; i++) {
		printf(" r%2d:%02x", i+ 16, gpr[i]);
	}
	printf("\n");
	for (; i<16; i++) {
		printf(" r%2d:%02x", i+ 16, gpr[i]);
	}
	printf("\n");
	printf("*** ram dump ***\n");
	addr = 0x40;
	for (n=0; n<16; n++) {
		if (!memcmp(ram+addr, ram+addr -16, 16)) {
			if (!match) {
				printf(" * \n");
			}
			match = 1;
			addr += 16;
			continue;
		}
		match = 0;
		printf("%04x:",addr);
		for (i=0; i<16; i++) {
			printf(" %02x",ram[addr+i]);
		}
		printf("\n");
		addr += 16;
	}
	printf("****************\n");
}

uint8_t avr_core_mem_read (AvrCore *core, int addr) {
	if (addr >= 0x4000) {
		addr -= 0x4000;
		addr &= (flash_size - 1);
		if (addr & 1) return (flash[addr >> 1] >> 8) & 0xff;
		return flash[addr >> 1] & 0xff;
	}
	if (addr >= 0x40) {
		addr &= (ram_size - 1);
		return ram[addr];
	}
	return avr_core_io_read (core, addr);
}

void avr_core_mem_write (AvrCore *core, int addr, uint8_t val) {
	if (addr >= 0x4000) {
		if (nvmcsr & 0x80) return;
		addr &= (flash_size - 1);
		if (!(addr & 1)) flash_tmp = val;
		else {
			flash[addr >> 1] = (uint16_t)val * 256 + flash_tmp;
		}
	}
	else if (addr >= 0x40) {
		addr &= (ram_size - 1);
		printf("S mem[%04x] = %02x : PC %04x FLAGS %02x SP %04x\n"
			, addr , val, PC , sreg.reg, SP);
		if (!(nvmcsr & 0x40) || ((addr & 0xff00) != 0x200))
			ram[addr] = val;
		log_done = 1;
	}
	else
		avr_core_io_write (core, addr, val);
}

uint8_t avr_core_sreg_get (AvrCore *core) {
	return sreg.reg;
}
void avr_core_sreg_set (AvrCore *core, uint8_t v) {
	sreg.reg = v;
}

int avr_core_sreg_get_bit (AvrCore *core, int b) {
	return (sreg.reg & (1 << b) ) != 0;
}

void avr_core_sreg_set_bit (AvrCore *core, int b, int v) {
	sreg.reg = set_bit_in_byte(sreg.reg, b, v);
}

uint8_t avr_core_gpwr_get (AvrCore *core, int reg) {
	return gpr[reg & 0xf];
}
void avr_core_gpwr_set (AvrCore *core, int reg, uint8_t val, int as_idx) {
	gpr[reg & 0xf] = val;
	if (!as_idx) {
		printf("L r%2d = %02x : PC %04x FLAGS %02x SP %04x\n"
			, reg , val, PC , sreg.reg, SP);
		log_done = 1;
	}
}

void avr_core_io_write (AvrCore *core, int addr, uint8_t val) {
	if (addr == 0x3f) sreg.reg = val;
	else if (addr == 0x3e) 
		SP = (SP & 0xff) | ((uint16_t)val << 8);
	else if (addr == 0x3d)
		SP = (SP & 0xff00) | (uint16_t)val;
	else if (addr == 0x33)  {
		nvmcmd = val & 0x3f;
		if (nvmcmd == 0x1d) nvmcsr &= ~0x80;
		else nvmcsr |= 0x80;
		if (nvmcmd == 0x2d) nvmcsr &= ~0x40;
		else nvmcsr |= 0x40;
	}
	else if (addr == 0x03)
		gpior0 = val;
	else if (addr == 0x07)
		gpior1 = val;
	else if (addr == 0x08)
		gpior2 = val;
	else if (addr == 0x1e)
		ram_addr_hi = val;
	else if (addr == 0x20)
		ram_addr = val;
	else if (addr == 0x1f) {
		addr = (ram_addr_hi << 8) | ram_addr;
		addr &= (ram_size - 1);
		ram[addr] = val;
		printf("S mem[%04x] = %02x : PC %04x FLAGS %02x SP %04x\n"
			, addr , val, PC , sreg.reg, SP);
	}
}

uint8_t avr_core_io_read (AvrCore *core, int addr) {
	if (addr == 0x3f) return sreg.reg;
	if (addr == 0x3e) return (SP >> 8) & 0xff;
	if (addr == 0x3d) return SP  & 0xff;
	if (addr == 0x32) return nvmcsr;
	if (addr == 0x33) return nvmcmd;
	if (addr == 0x03) return gpior0;
	if (addr == 0x07) return gpior1;
	if (addr == 0x08) return gpior2;
	if (addr == 0x20) return ram_addr;
	if (addr == 0x1e) return ram_addr_hi;
	if (addr == 0x1f) {
		addr = (ram_addr_hi << 8) | ram_addr;
		addr &= (ram_size - 1);
		return ram[addr];
	}
	return 0;
}

uint32_t avr_core_stack_pop (AvrCore *core, int bytes) {
	uint32_t r = 0;
	int i;
	for (i=0; i< bytes; i++) {
		r <<= 8;
		++ SP;
		r |= avr_core_mem_read (core, SP);
	}
	return r;
}

void avr_core_stack_push (AvrCore *core, int bytes, uint32_t val) {
	int i;
	for (i=0; i< bytes; i++) {
		SP --;
		avr_core_mem_write (core, SP + 1, val & 0xff);
		val >>= 8;
		PC++;
	}
	PC -= bytes;
}

int32_t avr_core_PC_size (AvrCore *core) {
	return 2;	// 2 bytes
}

int32_t avr_core_PC_get (AvrCore *core) {
	return PC;
}

void avr_core_PC_set (AvrCore *core, int32_t val) {
	PC = val;
}

void avr_core_PC_incr (AvrCore *core, int val) {
	PC = (PC + val) & 0xfff;
}

void avr_core_inst_CKS_set (AvrCore *core, int val) {
}

#  define strip_dir(path) (path)

void
private_avr_warning (char *file, int line, char *fmt, ...)
{
    va_list ap;
    char ffmt[128];

    snprintf (ffmt, sizeof (ffmt), "%s:%d: WARNING: %s", strip_dir (file),
              line, fmt);
    ffmt[127] = '\0';

    va_start (ap, fmt);
    vfprintf (stderr, ffmt, ap);
    va_end (ap);

    fflush (stderr);
}

void
private_avr_error (char *file, int line, char *fmt, ...)
{
    va_list ap;
    char ffmt[128];

    snprintf (ffmt, sizeof (ffmt), "\n%s:%d: ERROR: %s\n\n", strip_dir (file),
              line, fmt);
    ffmt[127] = '\0';

    va_start (ap, fmt);
    vfprintf (stderr, ffmt, ap);
    va_end (ap);

    fflush (stderr);

    exit (1);                   /* exit instead of abort */
}
