/* serjtag -- Serial Jtag Cable (using AVR) */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"
#include "cable.h"
#include "serjtag.h"
#include "serial.h"

#define SERJTAG_DEBUG		0
#define REQ_OUTSTANDINGS	10
static int req_count;

extern char * progname;
extern int do_cycles;
extern int ovsigck;
extern int verbose;

#define USE_INLINE_WRITE_PAGE
//#define USE_WRITE_PAGE_EEPROM
#define USE_INLINE_WRITE_PAGE_EEPROM

static int serjtag_send(CABLE cbl, char * buf, size_t len)
{
  return ser_send(&cbl->fd, (uint8_t *)buf, len);
}

static int serjtag_recv(CABLE cbl, char * buf, size_t len)
{
  int rv;

  rv = ser_recv(&cbl->fd, (uint8_t *)buf, len);
  if (rv < 0) {
    fprintf(stderr,
	    "%s: serjtag_recv(): programmer is not responding\n",
	    progname);
    exit(1);
  }
  return 0;
}

static int serjtag_drain(CABLE cbl, int display)
{
  char buf[128];
  int ret;

  ret = ser_drain(&cbl->fd, display);
  if (ret) return ret;
  memset(buf, 'V' , 3+JTAG_BUFSIZE);
  serjtag_send(cbl, buf, 3+JTAG_BUFSIZE);
  serjtag_recv(cbl, buf, 3+JTAG_BUFSIZE);
  return 0;
}

#if SERJTAG_DEBUG == 1
static void  serjtag_check(CABLE cbl) {
  int i,bytes;
  char buf[128];

fprintf(stderr,"check response\n");
fflush(stderr);
  i=0;
  buf[i++] = 'r'; 
  buf[i++] = 0;
  buf[i++] = 0;
  serjtag_send(cbl, buf, i);
  serjtag_recv(cbl, buf, 1);
fprintf(stderr,"\tCommonad = 0x%02x %c\n",buf[0],buf[0]);
  serjtag_recv(cbl, buf+1, 1);
fprintf(stderr,"\tFlags = 0x%02x\n",buf[1]);
  serjtag_recv(cbl, buf+2, 1);
fprintf(stderr,"\tBytes = 0x%02x\n",buf[2]);
  bytes = buf[2];
  if (buf[1] & JTAG_BITS) bytes++;
  if (bytes) {
	fprintf(stderr,"\t");
  	for (i=0; i<bytes; i++) {
  		serjtag_recv(cbl, buf+2+i, 1);
		fprintf(stderr,"%02x ",buf[2+i]);
  	}
	fprintf(stderr,"\n");
  }
}
#endif

#if 0
static int serjtag_recv_j(CABLE cbl, char * buf, size_t len, int verbose)
{
  int bytes;
  serjtag_recv(cbl, buf, 3);	/* header */
#if 0
  if (verbose) {
	fprintf(stderr,"recv_j(0) %c 0x%02x %d :",buf[0],buf[1],buf[2]);
  }
#endif
  bytes = buf[2];
  if (buf[1] & JTAG_BITS) {
	  bytes++;
  }
  len -= 3;
  if (len >= bytes) {
    serjtag_recv(cbl, buf+3, bytes);
  } else {
	char tmp[1];
	int i;
  	serjtag_recv(cbl, buf+3, len);
	for (i=0; i< bytes - len; i++) {
  		serjtag_recv(cbl, tmp, 1);
	}
  }
  if (verbose) {
	int i;
	fprintf(stderr,"recv_j %c 0x%02x %d :",buf[0],buf[1],buf[2]);
	for (i=0; i<bytes; i++) {
		fprintf(stderr,"%02x ",buf[3+i]&0xff);
	}
	fprintf(stderr,"\n");
  }
  return 0;
}
#endif

static int delay_param = 3;
static int use_delay = JTAG_USE_DELAY;

static int serjtag_set_delay(CABLE cbl, double bitclock) {
   use_delay = JTAG_USE_DELAY;
  if (bitclock == 0.0) { // using default
	delay_param = 3;
  } else if (bitclock >= 4.0) {
	delay_param = 0;
	use_delay = 0;
  } else {
  	delay_param =  (int)((2.26/3.0) / bitclock);
  	if (delay_param > 15) delay_param = 15;
  }
  if ((verbose>=1) || SERJTAG_DEBUG) {
	fprintf(stderr," serjtag:delay %d (%s) ( bitclk %.3f )\n"
		,delay_param, use_delay? "enabled":"disabled", bitclock);
  }
  return 0;
}

/*
 * initialize the CABLE and prepare it to accept commands
 */
static int serjtag_initialize(CABLE cbl)
{
  char id[8];
  char sw[3];
  char buf[3];

  /* Get the programmer identifier. Programmer returns exactly 7 chars
     _without_ the null.*/

  serjtag_send(cbl, "S", 1);
  serjtag_recv(cbl, id, 7);
  id[7] = 0;

  /* Get the HW and SW versions to see if the programmer is present. */

  serjtag_send(cbl, "V", 1);
  serjtag_recv(cbl, sw, 2);
  sw[2] = 0;

  fprintf(stderr, "Found programmer: Id = \"%s\"; Revison = %s\n", id, sw);

  if (strncmp(id, "SERJTAG", 7))
  {
     fprintf(stderr, "\tserjtag protocol not supported.\n");
     exit(1);
  }
  serjtag_set_delay(cbl, cbl->bitclock);

  serjtag_send(cbl, "j", 1); // enter jtag mode
  serjtag_recv(cbl, buf, 1);
  if (buf[0] != 'Y') {
	  return -1;
  }
  return 0;
}

static int serjtag_open(CABLE cbl, char * port)
{
  int ret;
  /*
   *  If baudrate was not specified use 115.200 Baud
   */
  if(cbl->baudrate == 0) {
    cbl->baudrate = 115200;
  }

  //strcpy(cbl->port, port);
  ser_open(port, cbl->baudrate, &cbl->fd);
  /*
   * drain any extraneous input
   */
  ser_set_dtr_rts(&cbl->fd, 0);
  usleep(50*1000);
  serjtag_drain (cbl, 0);
	
  ret = serjtag_initialize(cbl);
  if (ret) return ret;
  return ret;
}

static void serjtag_close(CABLE cbl, int do_reset)
{
  char buf[2];
  // exit jtag mode
  serjtag_send(cbl, "V", 1);
  serjtag_recv(cbl, buf, 2);

  ser_close(&cbl->fd);
  cbl->fd.ifd = -1;
}

static struct request {
	int bytes;
	uint8_t *buf;
	int n_bits;
	struct request *next;
} *req_head,*req_tail,*req_pool;

static void put_request(int bytes, uint8_t *buf, int n_bits)
{
  struct request *p;
  if (req_pool) {
     p = req_pool;
     req_pool = p->next;
  } else {
     p = malloc(sizeof(struct request));
     if (!p) {
       fprintf(stderr, "can't alloc memory\n");
       exit(1);
     }
  }
  memset(p, 0, sizeof(struct request));
  p->bytes = bytes;
  p->buf = buf;
  p->n_bits = n_bits;
  if (req_tail) {
     req_tail->next = p;
     req_tail = p;
  } else {
     req_head = req_tail = p;
  }
}

static int do_request(CABLE cbl)
{
  struct request *p;
  int bytes,i;
  uint8_t *rcv_buf;
  int n_bits;
  uint8_t bb_buf[128];

  if (!req_head) return 0;
  p = req_head;
  req_head = p->next;
  if (!req_head) req_tail = req_head;

  bytes = p->bytes;
  rcv_buf = p->buf;
  n_bits = p->n_bits;

  memset(p, 0, sizeof(struct request));
  p->next = req_pool;
  req_pool = p;

  serjtag_recv(cbl, bb_buf, 3+bytes);
  for (i=0; i<bytes; i++) {
	rcv_buf[i] = rev[bb_buf[3+i]];
  }
  return 1;
}

static int put_tdi_bits(CABLE cbl, uint8_t *buf, int off, int bit_len,
				    int flags, uint8_t *rcv_buf, int roff) 
{
  uint32_t    i,j,n;
  int tms_high,recive;
  int buf_pos = 0;
  uint8_t bb_buf[128];

  tms_high = 0;
  recive = flags & CABLE_RECIEVE? JTAG_RECIEVE : 0;
  for (i=0; i<bit_len; ) {
     buf_pos = 0;
     n = bit_len - i;
     if (n > JTAG_BUFSIZE * 8) n = JTAG_BUFSIZE * 8;
     else {
	if (flags & CABLE_TMS_HIGH) tms_high = JTAG_TMS_HIGH;
     }
     bb_buf[buf_pos++] = JTAG_CMD_PUT_TDI_STREAM; // CMD 'd'
     bb_buf[buf_pos++] = (n & JTAG_BITS) | tms_high | recive;
     bb_buf[buf_pos++] = (n >> 3);
     j = (n + 7) >> 3;
     for (i=0; i<j; i++) {
         bb_buf[buf_pos++] = rev[buf[i]];
     }
     if (flags & CABLE_RECIEVE) {
	bb_buf[buf_pos++] = JTAG_CMD_REQUEST_DATA; // CMD 'r'
	bb_buf[buf_pos++] = (n & JTAG_BITS); // dummy
	bb_buf[buf_pos++] = (n >>3); // dummy
	put_request(j, rcv_buf, n);
	rcv_buf += j;
     }
     serjtag_send(cbl, bb_buf, buf_pos);
     buf += j;
     i += n;
     req_count++;
     if (req_count > REQ_OUTSTANDINGS)
        do_request(cbl);
  }
  while (do_request(cbl))
	  ;
  return bit_len;
}

static int put_tms_tdi_bits(CABLE cbl, uint8_t *buf, int off, int bit_len,
				    int flags, uint8_t *rcv_buf, int roff) 
{
  uint32_t    i,j,n;
  int recive;
  int buf_pos = 0;
  uint8_t bb_buf[128];

  recive = flags & CABLE_RECIEVE? JTAG_RECIEVE : 0;
  for (i=0; i<bit_len; ) {
     buf_pos = 0;
     n = bit_len - i;
     if (n > JTAG_BUFSIZE * 4) n = JTAG_BUFSIZE * 4;
     bb_buf[buf_pos++] = JTAG_CMD_PUT_TMS_TDI_STREAM; // CMD 'D'
     bb_buf[buf_pos++] = (n*2 & JTAG_BITS) | recive;
     bb_buf[buf_pos++] = (n*2 >> 3);
     j = (n*2 + 7) >> 3;
     for (i=0; i<j; i++) {
         bb_buf[buf_pos++] = rev[buf[i]];
     }
     if (flags & CABLE_RECIEVE) {
	bb_buf[buf_pos++] = JTAG_CMD_REQUEST_DATA; // CMD 'r'
	bb_buf[buf_pos++] = (n & JTAG_BITS); // dummy
	bb_buf[buf_pos++] = (n >>3); // dummy
	put_request((n+7)/8, rcv_buf, n);
	rcv_buf += (n+7)/8;
     }
     serjtag_send(cbl, bb_buf, buf_pos);
     buf += j;
     i += n;
     req_count++;
     if (req_count > REQ_OUTSTANDINGS)
        do_request(cbl);
  }
  while (do_request(cbl))
	  ;
  return bit_len;
}

#if 0
static int get_tdo_bits(CABLE cbl, int bit_len, int tdi, uint8_t *rcv_buf)
{
  uint32_t    i,j,n;
  int tdi_high,recive;
  int buf_pos = 0;
  uint8_t bb_buf[10];

  tdi_high = (tdi) ? JTAG_TDI_HIGH : 0;
  recive = JTAG_RECIEVE;
  for (i=0; i<bit_len; ) {
     buf_pos = 0;
     n = bit_len - i;
     if (n > JTAG_BUFSIZE * 8) n = JTAG_BUFSIZE * 8;
     bb_buf[buf_pos++] = JTAG_CMD_GET_TDO_STREAM; // CMD 'c'
     bb_buf[buf_pos++] = (n & JTAG_BITS) | tdi_high | recive;
     bb_buf[buf_pos++] = (n >> 3);
     j = (n + 7) >> 3;
     bb_buf[buf_pos++] = JTAG_CMD_REQUEST_DATA; // CMD 'r'
     bb_buf[buf_pos++] = (n & JTAG_BITS); // dummy
     bb_buf[buf_pos++] = (n >>3); // dummy
     serjtag_send(cbl, bb_buf, buf_pos);
     put_request(j, rcv_buf, n);
     rcv_buf += j;
     i += n;
     req_count++;
     if (req_count > REQ_OUTSTANDINGS)
        do_request(cbl);
  }
  while (do_request(cbl))
	  ;
  return bit_len;
}
#endif

static int setup_port(CABLE cbl, int tms, int tdi, int tck) {
  uint8_t buf[10];
  int i;
  i=0;
  buf[i++] = 's';		/* Set Port */
  buf[i++] = 0;			/* flags */
  buf[i++] = 1;			/* bytes */
  buf[i++] = (tms ? JTAG_SET_TMS : 0 )
  	   | (tdi ? JTAG_SET_TDI : 0 )
  	   | (tck ? JTAG_SET_TCK : 0 )
	   | delay_param;

  serjtag_send(cbl, buf, i);
  serjtag_recv(cbl, buf, i);
  return 0;
}

static int get_gpio(CABLE cbl, int bit) {
	return 0;
}

static int setup_gpio(CABLE cbl, int bit, int value) {
	return 0;
}

void serjtag_init(CABLE cbl)
{
  /*
   * functions
   */
  cbl->open               = serjtag_open;
  cbl->close              = serjtag_close;

  cbl->setup_port         = setup_port;
  cbl->put_tdi_bits       = put_tdi_bits;
  cbl->put_tms_tdi_bits   = put_tms_tdi_bits;

  cbl->setup_gpio         = setup_gpio;
  cbl->get_gpio           = get_gpio;

//  cbl->set_bitclock       = set_bitclock;
//  cbl->parseexitspecs = parseexitspecs;

}
